@@ -1,11 +1,19 @@
 /*
  * ModSecurity for Apache 2.x, http://www.modsecurity.org/
- * Copyright (c) 2004-2008 Breach Security, Inc. (http://www.breach.com/)
+ * Copyright (c) 2004-2009 Breach Security, Inc. (http://www.breach.com/)
  *
- * You should have received a copy of the licence along with this
- * program (stored in the file "LICENSE"). If the file is missing,
- * or if you have any other questions related to the licence, please
- * write to Breach Security, Inc. at support@breach.com.
+ * This product is released under the terms of the General Public Licence,
+ * version 2 (GPLv2). Please refer to the file LICENSE (included with this
+ * distribution) which contains the complete text of the licence.
+ *
+ * There are special exceptions to the terms and conditions of the GPL
+ * as it is applied to this software. View the full text of the exception in
+ * file MODSECURITY_LICENSING_EXCEPTION in the directory of this software
+ * distribution.
+ *
+ * If any of the files related to licensing are missing or if you have any
+ * other questions related to licensing please contact Breach Security, Inc.
+ * directly using the email address support@breach.com.
  *
  */
 #include <limits.h>
@@ -15,6 +23,9 @@
 
 #include "modsecurity.h"
 #include "apache2.h"
+#include "http_main.h"
+#include "pdf_protect.h"
+
 #include "msc_logging.h"
 #include "msc_util.h"
 
@@ -23,7 +34,6 @@
 
 msc_engine DSOLOCAL *modsecurity = NULL;
 
-
 /* Global module variables; these are used for the Apache-specific functionality */
 
 char DSOLOCAL *chroot_dir = NULL;
@@ -63,7 +73,7 @@ int perform_interception(modsec_rec *msr) {
     }
 
     if (msr->phase > 4) {
-        msr_log(msr, 1, "Internal Error: Asked to intercept request in phase %i.", msr->phase);
+        msr_log(msr, 1, "Internal Error: Asked to intercept request in phase %d.", msr->phase);
         msr->was_intercepted = 0;
         return DECLINED;
     }
@@ -71,7 +81,7 @@ int perform_interception(modsec_rec *msr) {
     /* OK, we're good to go. */
 
     actionset = msr->intercept_actionset;
-    phase_text = apr_psprintf(msr->mp, " (phase %i)", msr->phase);
+    phase_text = apr_psprintf(msr->mp, " (phase %d)", msr->phase);
 
     /* By default we log at level 1 but we switch to 4
      * if a nolog action was used or this is not the initial request
@@ -82,7 +92,7 @@ int perform_interception(modsec_rec *msr) {
     /* Pause the request first (if configured and the initial request). */
     if (actionset->intercept_pause) {
         msr_log(msr, (log_level > 3 ? log_level : log_level + 1), "Pausing transaction for "
-            "%i msec.", actionset->intercept_pause);
+            "%d msec.", actionset->intercept_pause);
         /* apr_sleep accepts microseconds */
         apr_sleep((apr_interval_time_t)(actionset->intercept_pause * 1000));
     }
@@ -92,13 +102,13 @@ int perform_interception(modsec_rec *msr) {
         case ACTION_DENY :
             if (actionset->intercept_status != 0) {
                 status = actionset->intercept_status;
-                message = apr_psprintf(msr->mp, "Access denied with code %i%s.",
+                message = apr_psprintf(msr->mp, "Access denied with code %d%s.",
                     status, phase_text);
             } else {
                 log_level = 1;
                 status = HTTP_INTERNAL_SERVER_ERROR;
                 message = apr_psprintf(msr->mp, "Access denied with code 500%s "
-                    "(Internal Error: Invalid status code requested %i).",
+                    "(Internal Error: Invalid status code requested %d).",
                     phase_text, actionset->intercept_status);
             }
             break;
@@ -139,7 +149,7 @@ int perform_interception(modsec_rec *msr) {
                 extern module core_module;
                 apr_socket_t *csd = ap_get_module_config(msr->r->connection->conn_config,
                     &core_module);
-                
+
                 if (csd) {
                     if (apr_socket_close(csd) == APR_SUCCESS) {
                         status = HTTP_FORBIDDEN;
@@ -180,7 +190,7 @@ int perform_interception(modsec_rec *msr) {
                 status = HTTP_MOVED_TEMPORARILY;
             }
             message = apr_psprintf(msr->mp, "Access denied with redirection to %s using "
-                "status %i%s.",
+                "status %d%s.",
                 log_escape_nq(msr->mp, actionset->intercept_uri), status,
                 phase_text);
             break;
@@ -189,13 +199,28 @@ int perform_interception(modsec_rec *msr) {
             status = DECLINED;
             message = apr_psprintf(msr->mp, "Access allowed%s.", phase_text);
             msr->was_intercepted = 0;
+            msr->allow_scope = ACTION_ALLOW;
+            break;
+
+        case ACTION_ALLOW_PHASE :
+            status = DECLINED;
+            message = apr_psprintf(msr->mp, "Access to phase allowed%s.", phase_text);
+            msr->was_intercepted = 0;
+            msr->allow_scope = ACTION_ALLOW_PHASE;
+            break;
+
+        case ACTION_ALLOW_REQUEST :
+            status = DECLINED;
+            message = apr_psprintf(msr->mp, "Access to request allowed%s.", phase_text);
+            msr->was_intercepted = 0;
+            msr->allow_scope = ACTION_ALLOW_REQUEST;
             break;
 
         default :
             log_level = 1;
             status = HTTP_INTERNAL_SERVER_ERROR;
             message = apr_psprintf(msr->mp, "Access denied with code 500%s "
-                "(Internal Error: invalid interception action %i).",
+                "(Internal Error: invalid interception action %d).",
                 phase_text, actionset->intercept_action);
             break;
     }
@@ -328,18 +353,21 @@ static modsec_rec *create_tx_context(request_rec *r) {
     msr->request_headers = apr_table_copy(msr->mp, r->headers_in);
     msr->hostname = ap_get_server_name(r);
 
+    msr->msc_rule_mptmp = NULL;
+
     /* Invoke the engine to continue with initialisation */
-    if (modsecurity_tx_init(msr) < 0) return NULL;
+    if (modsecurity_tx_init(msr) < 0) {
+        msr_log(msr, 1, "Failed to initialise transaction (txid %s).", msr->txid);
+        return NULL;
+    }
 
     store_tx_context(msr, r);
 
     if (msr->txcfg->debuglog_level >= 4) {
-        msr_log(msr, 4, "Transaction context created (dcfg %x).", msr->dcfg1);
+        msr_log(msr, 4, "Transaction context created (dcfg %pp).", msr->dcfg1);
     }
 
-    msr->msc_rule_mptmp = NULL;
-
-    return msr;    
+    return msr;
 }
 
 
@@ -355,7 +383,8 @@ static apr_status_t change_server_signature(server_rec *s) {
 
     if (new_server_signature == NULL) return 0;
 
-    server_version = (char *)ap_get_server_version();
+    server_version = (char *)apache_get_server_version();
+
     if (server_version == NULL) {
         ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
             "SecServerSignature: Apache returned null as signature.");
@@ -372,6 +401,18 @@ static apr_status_t change_server_signature(server_rec *s) {
         return -1;
     }
 
+    /* Check that it really changed.  This assumes that what we got was
+     * not a copy and this could change in future versions of Apache.
+     */
+    server_version = (char *)apache_get_server_version();
+    if ((server_version == NULL) || (strcmp(server_version, new_server_signature) != 0)) {
+        ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s, "SecServerSignature: Failed to change server signature to \"%s\".", new_server_signature);
+        return 0;
+    }
+    else {
+        ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, s, "SecServerSignature: Changed server signature to \"%s\".", server_version);
+    }
+
     return 1;
 }
 
@@ -390,12 +431,10 @@ static int hook_pre_config(apr_pool_t *mp, apr_pool_t *mp_log, apr_pool_t *mp_te
     /* Initialise ModSecurity engine */
     modsecurity = modsecurity_create(mp, MODSEC_ONLINE);
     if (modsecurity == NULL) {
-        /* ENH Since s not available, how do we log from here? stderr?
-         * ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
-         *    "ModSecurity: Failed to initialise engine.");
-         */
+        ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL,
+            "ModSecurity: Failed to initialise engine.");
         return HTTP_INTERNAL_SERVER_ERROR;
-    }    
+    }
 
     return OK;
 }
@@ -407,6 +446,10 @@ static int hook_post_config(apr_pool_t *mp, apr_pool_t *mp_log, apr_pool_t *mp_t
     void *init_flag = NULL;
     int first_time = 0;
 
+    /* ENH Figure out a way to validate config before we start
+     * - skipafter: need to make sure we found all of our targets
+     */
+
     /* Figure out if we are here for the first time */
     apr_pool_userdata_get(&init_flag, "modsecurity-init-flag", s->process->pool);
     if (init_flag == NULL) {
@@ -418,7 +461,7 @@ static int hook_post_config(apr_pool_t *mp, apr_pool_t *mp_log, apr_pool_t *mp_t
     }
 
     /* Store the original server signature */
-    real_server_signature = apr_pstrdup(mp, ap_get_server_version());
+    real_server_signature = apr_pstrdup(mp, apache_get_server_version());
 
     /* Make some space in the server signature for later */
     if (new_server_signature != NULL) {
@@ -438,7 +481,7 @@ static int hook_post_config(apr_pool_t *mp, apr_pool_t *mp_log, apr_pool_t *mp_t
 
         if (first_time == 0) {
             ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
-                "ModSecurity: chroot checkpoint #2 (pid=%d ppid=%d)", (int)getpid(), (int)getppid());
+                "ModSecurity: chroot checkpoint #2 (pid=%ld ppid=%ld)", (long)getpid(), (long)getppid());
 
             if (chdir(chroot_dir) < 0) {
                 ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
@@ -467,7 +510,7 @@ static int hook_post_config(apr_pool_t *mp, apr_pool_t *mp_log, apr_pool_t *mp_t
                 "ModSecurity: chroot successful, path=%s", chroot_dir);
         } else {
             ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
-                "ModSecurity: chroot checkpoint #1 (pid=%d ppid=%d)", (int)getpid(), (int)getppid());
+                "ModSecurity: chroot checkpoint #1 (pid=%ld ppid=%ld)", (long)getpid(), (long)getppid());
         }
     }
     #endif
@@ -477,13 +520,13 @@ static int hook_post_config(apr_pool_t *mp, apr_pool_t *mp_log, apr_pool_t *mp_t
 
     /* Log our presence to the error log. */
     if (first_time) {
+        ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
+                "%s configured.", MODSEC_MODULE_NAME_FULL);
+
+        /* If we've changed the server signature make note of the original. */
         if (new_server_signature != NULL) {
             ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
-                "ModSecurity for Apache %s configured - %s", MODULE_RELEASE, real_server_signature);
-        }
-        else {
-            ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
-                "ModSecurity for Apache %s configured", MODULE_RELEASE);
+                "Original server signature: %s", real_server_signature);
         }
     }
 
@@ -528,19 +571,24 @@ static int hook_request_early(request_rec *r) {
         return DECLINED;
     }
 
-    /* Check request body limit (should only trigger on non-chunked requests). */
-    if (msr->request_content_length > msr->txcfg->reqbody_limit) {
-        msr_log(msr, 1, "Request body is larger than the "
-                     "configured limit (%lu).", msr->txcfg->reqbody_limit);
-        return HTTP_REQUEST_ENTITY_TOO_LARGE;
-    }
-
     /* Process phase REQUEST_HEADERS */
     rc = DECLINED;
     if (modsecurity_process_phase(msr, PHASE_REQUEST_HEADERS) > 0) {
         rc = perform_interception(msr);
     }
 
+    if (    (msr->txcfg->is_enabled != MODSEC_DISABLED)
+         && (msr->txcfg->reqbody_access == 1)
+         && (rc == DECLINED))
+    {
+        /* Check request body limit (non-chunked requests only). */
+        if (msr->request_content_length > msr->txcfg->reqbody_limit) {
+            msr_log(msr, 1, "Request body (Content-Length) is larger than the "
+                         "configured limit (%ld).", msr->txcfg->reqbody_limit);
+            return HTTP_REQUEST_ENTITY_TOO_LARGE;
+        }
+    }
+
     return rc;
 }
 
@@ -558,7 +606,7 @@ static int hook_request_late(request_rec *r) {
      */
     if ((r->main != NULL)||(r->prev != NULL)) {
         return DECLINED;
-    } 
+    }
 
     /* Find the transaction context and make sure
      * we are supposed to proceed.
@@ -582,7 +630,7 @@ static int hook_request_late(request_rec *r) {
 
     /* Get the second configuration context. */
     msr->dcfg2 = (directory_config *)ap_get_module_config(r->per_dir_config,
-        &security2_module);    
+        &security2_module);
 
     /* Create a transaction context. */
     msr->txcfg = create_directory_config(msr->mp, NULL);
@@ -597,6 +645,15 @@ static int hook_request_late(request_rec *r) {
 
     init_directory_config(msr->txcfg);
 
+    /* Perform the PDF tests now. */
+    rc = pdfp_check(msr);
+    if (rc > 0) {
+        /* The PDF protection module has decided it needs to
+         * redirect the current transaction. So we let it do that.
+         */
+        return rc;
+    }
+
     if (msr->txcfg->is_enabled == 0) {
         if (msr->txcfg->debuglog_level >= 4) {
             msr_log(msr, 4, "Processing disabled, skipping (hook request_late).");
@@ -605,7 +662,7 @@ static int hook_request_late(request_rec *r) {
     }
 
     if (msr->txcfg->debuglog_level >= 4) {
-        msr_log(msr, 4, "Second phase starting (dcfg %x).", msr->dcfg2);
+        msr_log(msr, 4, "Second phase starting (dcfg %pp).", msr->dcfg2);
     }
 
     /* Figure out whether or not to extract multipart files. */
@@ -620,14 +677,22 @@ static int hook_request_late(request_rec *r) {
     if (rc < 0) {
         switch(rc) {
             case -1 :
-                msr_log(msr, 1, "%s", my_error_msg);
+                if (my_error_msg != NULL) {
+                    msr_log(msr, 1, "%s", my_error_msg);
+                }
                 return HTTP_INTERNAL_SERVER_ERROR;
                 break;
             case -4 : /* Timeout. */
+                if (my_error_msg != NULL) {
+                    msr_log(msr, 4, "%s", my_error_msg);
+                }
                 r->connection->keepalive = AP_CONN_CLOSE;
                 return HTTP_REQUEST_TIME_OUT;
                 break;
             case -5 : /* Request body limit reached. */
+                if (my_error_msg != NULL) {
+                    msr_log(msr, 1, "%s", my_error_msg);
+                }
                 r->connection->keepalive = AP_CONN_CLOSE;
                 return HTTP_REQUEST_ENTITY_TOO_LARGE;
                 break;
@@ -643,8 +708,9 @@ static int hook_request_late(request_rec *r) {
     /* Update the request headers. They might have changed after
      * the body was read (trailers).
      */
-    // TODO We still need to keep a copy of the original headers
-    //      to log in the audit log.
+    /* NOTE We still need to keep a copy of the original headers
+     *      to log in the audit log.
+     */
     msr->request_headers = apr_table_copy(msr->mp, r->headers_in);
 
     /* Process phase REQUEST_BODY */
@@ -672,12 +738,12 @@ static void hook_error_log(const char *file, int line, int level, apr_status_t s
     if (r == NULL) return;
     msr = retrieve_tx_context((request_rec *)r);
 
-    /* Create a context for requests we never had the chance to process */ 
+    /* Create a context for requests we never had the chance to process */
     if ((msr == NULL)
         && ((level & APLOG_LEVELMASK) < APLOG_DEBUG)
         && apr_table_get(r->subprocess_env, "UNIQUE_ID"))
     {
-        msr = create_tx_context((request_rec *)r); 
+        msr = create_tx_context((request_rec *)r);
         if (msr->txcfg->debuglog_level >= 9) {
             if (msr == NULL) {
                 msr_log(msr, 9, "Failed to create context after request failure.");
@@ -686,7 +752,7 @@ static void hook_error_log(const char *file, int line, int level, apr_status_t s
                 msr_log(msr, 9, "Context created after request failure.");
             }
         }
-    } 
+    }
 
     if (msr == NULL) return;
 
@@ -753,7 +819,7 @@ static void sec_guardian_logger(request_rec *r, request_rec *origr, modsec_rec *
      * The fields SESSION_ID, MODSEC_MESSAGE, and MODSEC_RATING are not used at the moment.
      */
 
-    str2 = apr_psprintf(msr->mp, "%" APR_TIME_T_FMT " %" APR_TIME_T_FMT " \"%s\" %i",
+    str2 = apr_psprintf(msr->mp, "%" APR_TIME_T_FMT " %" APR_TIME_T_FMT " \"%s\" %d",
         duration, apr_time_sec(duration), log_escape(msr->mp, modsec_message), modsec_rating);
     if (str2 == NULL) return;
 
@@ -774,7 +840,7 @@ static void sec_guardian_logger(request_rec *r, request_rec *origr, modsec_rec *
 
     limit = limit - strlen(str2) - 5;
     if (limit <= 0) {
-        msr_log(msr, 1, "Audit Log: Atomic PIPE write buffer too small: %i", PIPE_BUF);
+        msr_log(msr, 1, "Audit Log: Atomic PIPE write buffer too small: %d", PIPE_BUF);
         return;
     }
 
@@ -829,11 +895,11 @@ static int hook_log_transaction(request_rec *r) {
     while ((arr->nelts == 0)&&(r->prev != NULL)) {
         r = r->prev;
         arr = apr_table_elts(r->headers_out);
-    }    
+    }
 
     msr->r = r;
     msr->response_status = r->status;
-    msr->status_line = ((r->status_line != NULL)    
+    msr->status_line = ((r->status_line != NULL)
         ? r->status_line : ap_get_status_line(r->status));
     msr->response_protocol = get_response_protocol(origr);
     msr->response_headers = apr_table_copy(msr->mp, r->headers_out);
@@ -867,13 +933,13 @@ static void hook_insert_filter(request_rec *r) {
     /* Add the input filter, but only if we need it to run. */
     if (msr->if_status == IF_STATUS_WANTS_TO_RUN) {
         if (msr->txcfg->debuglog_level >= 4) {
-            msr_log(msr, 4, "Hook insert_filter: Adding input forwarding filter %s(r %x).", (((r->main != NULL)||(r->prev != NULL)) ? "for subrequest " : ""), r);
+            msr_log(msr, 4, "Hook insert_filter: Adding input forwarding filter %s(r %pp).",
+                (((r->main != NULL)||(r->prev != NULL)) ? "for subrequest " : ""), r);
         }
 
         ap_add_input_filter("MODSECURITY_IN", msr, r, r->connection);
     }
 
-
     /* The output filters only need to be added only once per transaction
      * (i.e. subrequests and redirects are excluded).
      */
@@ -881,7 +947,14 @@ static void hook_insert_filter(request_rec *r) {
         return;
     }
 
-    /* Only proceed to add the filter if the engine is enabled. */
+    /* We always add the PDF XSS protection filter. */
+    if (msr->txcfg->debuglog_level >= 4) {
+        msr_log(msr, 4, "Hook insert_filter: Adding PDF XSS protection output filter (r %pp).", r);
+    }
+
+    ap_add_output_filter("PDFP_OUT", msr, r, r->connection);
+
+    /* Only proceed to add the second filter if the engine is enabled. */
     if (msr->txcfg->is_enabled == 0) {
         if (msr->txcfg->debuglog_level >= 4) {
             msr_log(msr, 4, "Hook insert_filter: Processing disabled, skipping.");
@@ -896,13 +969,14 @@ static void hook_insert_filter(request_rec *r) {
      */
     if (msr->of_status != OF_STATUS_COMPLETE) {
         if (msr->txcfg->debuglog_level >= 4) {
-            msr_log(msr, 4, "Hook insert_filter: Adding output filter (r %x).", r);
+            msr_log(msr, 4, "Hook insert_filter: Adding output filter (r %pp).", r);
         }
 
         ap_add_output_filter("MODSECURITY_OUT", msr, r, r->connection);
     }
 }
 
+// TODO: Holding off on this for now (needs more testing)
 #if 0
 /**
  * Invoked whenever Apache starts processing an error. A chance
@@ -914,25 +988,30 @@ static void hook_insert_error_filter(request_rec *r) {
     /* Find the transaction context and make sure we are
      * supposed to proceed.
      */
-
-    /* TODO Insert filter but make a note that it's the error
-     *      response the filter would be receiving.
-     */
-
     msr = retrieve_tx_context(r);
     if (msr == NULL) return;
 
+    /* Do not run if not enabled. */
     if (msr->txcfg->is_enabled == 0) {
         if (msr->txcfg->debuglog_level >= 4) {
             msr_log(msr, 4, "Hook insert_error_filter: Processing disabled, skipping.");
         }
         return;
     }
-    
+
+    /* Do not run if the output filter already completed. This will
+     * happen if we intercept in phase 4.
+     */
     if (msr->of_status != OF_STATUS_COMPLETE) {
         if (msr->txcfg->debuglog_level >= 4) {
-            msr_log(msr, 4, "Hook insert_error_filter: Adding output filter (r %x).", r);
+            msr_log(msr, 4, "Hook insert_error_filter: Adding output filter (r %pp).", r);
         }
+
+        /* Make a note that the output we will be receiving is a
+         * result of error processing.
+         */
+        msr->of_is_error = 1;
+
         ap_add_output_filter("MODSECURITY_OUT", msr, r, r->connection);
     } else {
         if (msr->txcfg->debuglog_level >= 4) {
@@ -949,7 +1028,7 @@ static void hook_insert_error_filter(request_rec *r) {
  */
 static void modsec_register_tfn(const char *name, void *fn) {
     if (modsecurity != NULL) {
-        msre_engine_tfn_register(modsecurity->msre, name, fn);
+        msre_engine_tfn_register(modsecurity->msre, name, (fn_tfn_execute_t)fn);
     }
 }
 
@@ -959,7 +1038,23 @@ static void modsec_register_tfn(const char *name, void *fn) {
  */
 static void modsec_register_operator(const char *name, void *fn_init, void *fn_exec) {
     if (modsecurity != NULL) {
-        msre_engine_op_register(modsecurity->msre, name, fn_init, fn_exec);
+        msre_engine_op_register(modsecurity->msre, name, (fn_op_param_init_t)fn_init, (fn_op_execute_t)fn_exec);
+    }
+}
+
+/**
+ * This function is exported for other Apache modules to
+ * register new variables.
+ */
+static void modsec_register_variable(const char *name, unsigned int type,
+                                     unsigned int argc_min, unsigned int argc_max,
+                                     void *fn_validate, void *fn_generate,
+                                     unsigned int is_cacheable, unsigned int availability) {
+    if (modsecurity != NULL) {
+        msre_engine_variable_register(modsecurity->msre, name, type, argc_min, argc_max, fn_validate, fn_generate, is_cacheable, availability);
+    }
+    else {
+        fprintf(stderr,"modsecurity is NULL\n");
     }
 }
 #endif
@@ -980,7 +1075,9 @@ static void register_hooks(apr_pool_t *mp) {
     };
     static const char *postread_beforeme_list[] = {
         "mod_rpaf.c",
+        "mod_rpaf-2.0.c",
         "mod_extract_forwarded2.c",
+        "mod_custom_header.c",
         "mod_breach_realip.c",
         "mod_breach_trans.c",
         "mod_unique_id.c",
@@ -991,10 +1088,14 @@ static void register_hooks(apr_pool_t *mp) {
         NULL
     };
 
+    /* Add the MODSEC_a.b define */
+    *(char **)apr_array_push(ap_server_config_defines) = apr_psprintf(mp, "MODSEC_%s.%s", MODSEC_VERSION_MAJOR, MODSEC_VERSION_MINOR);
+
 #if (!defined(NO_MODSEC_API))
     /* Export optional functions. */
     APR_REGISTER_OPTIONAL_FN(modsec_register_tfn);
     APR_REGISTER_OPTIONAL_FN(modsec_register_operator);
+    APR_REGISTER_OPTIONAL_FN(modsec_register_variable);
 #endif
 
     /* Main hooks */
@@ -1006,7 +1107,7 @@ static void register_hooks(apr_pool_t *mp) {
     /* Our own hook to handle RPC transactions (not used at the moment).
      * // ap_hook_handler(hook_handler, NULL, NULL, APR_HOOK_MIDDLE);
      */
-    
+
     /* Transaction processing hooks */
     ap_hook_post_read_request(hook_request_early,
         postread_beforeme_list, postread_afterme_list, APR_HOOK_REALLY_FIRST);
@@ -1019,12 +1120,17 @@ static void register_hooks(apr_pool_t *mp) {
 
     /* Filter hooks */
     ap_hook_insert_filter(hook_insert_filter, NULL, NULL, APR_HOOK_FIRST);
-    /* ap_hook_insert_error_filter(hook_insert_error_filter, NULL, NULL, APR_HOOK_FIRST); */
+#if 0
+    ap_hook_insert_error_filter(hook_insert_error_filter, NULL, NULL, APR_HOOK_FIRST);
+#endif
 
     ap_register_input_filter("MODSECURITY_IN", input_filter,
         NULL, AP_FTYPE_CONTENT_SET);
     ap_register_output_filter("MODSECURITY_OUT", output_filter,
         NULL, AP_FTYPE_CONTENT_SET);
+
+    ap_register_output_filter("PDFP_OUT", pdfp_output_filter,
+        NULL, AP_FTYPE_CONTENT_SET);
 }
 
 /* Defined in apache2_config.c */
