@@ -1,11 +1,19 @@
 /*
  * ModSecurity for Apache 2.x, http://www.modsecurity.org/
- * Copyright (c) 2004-2008 Breach Security, Inc. (http://www.breach.com/)
+ * Copyright (c) 2004-2009 Breach Security, Inc. (http://www.breach.com/)
  *
- * You should have received a copy of the licence along with this
- * program (stored in the file "LICENSE"). If the file is missing,
- * or if you have any other questions related to the licence, please
- * write to Breach Security, Inc. at support@breach.com.
+ * This product is released under the terms of the General Public Licence,
+ * version 2 (GPLv2). Please refer to the file LICENSE (included with this
+ * distribution) which contains the complete text of the licence.
+ *
+ * There are special exceptions to the terms and conditions of the GPL
+ * as it is applied to this software. View the full text of the exception in
+ * file MODSECURITY_LICENSING_EXCEPTION in the directory of this software
+ * distribution.
+ *
+ * If any of the files related to licensing are missing or if you have any
+ * other questions related to licensing please contact Breach Security, Inc.
+ * directly using the email address support@breach.com.
  *
  */
 #ifndef _MSC_LOGGING_H_
@@ -28,7 +36,9 @@
 #define AUDITLOG_PART_A_RESPONSE_BODY       'G'
 #define AUDITLOG_PART_TRAILER               'H'
 #define AUDITLOG_PART_FAKE_REQUEST_BODY     'I'
-#define AUDITLOG_PART_LAST                  'I'
+#define AUDITLOG_PART_UPLOADS               'J'
+#define AUDITLOG_PART_MATCHEDRULES          'K'
+#define AUDITLOG_PART_LAST                  'K'
 #define AUDITLOG_PART_ENDMARKER             'Z'
 
 #include "modsecurity.h"
